#!/usr/local/bin/ruby19
# coding:utf-8
# vi:set ts=3 sw=3:
# vim:set sts=0 noet:

=begin
Copyright (c) 2010 KISHIMOTO, Makoto

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
=end

def output_script_1 io
	io.write <<'END_OF_SCRIPT'
#!/usr/local/bin/ruby19
# coding:utf-8
# vi:set ts=3 sw=3:
# vim:set sts=0 noet:

require 'stringio'

DEBUG_DUMP = false
#DEBUG_DUMP = true

class GifCat
	def initialize
		@gifs = []
	end

	def concat gifdata
		info = GifInfo.new gifdata
		@gifs << info
	end

	def get_content
		width, height = 0, 0
		useLocalColorTable = false
		globalColorTable = nil
		@gifs.each do|info|
			width += info.screenWidth
			if info.screenHeight > height then
				height = info.screenHeight
			end
			unless useLocalColorTable then
				if globalColorTable then
					if info.globalColorTableFlag then
						if globalColorTable != info.globalColorTable then
							useLocalColorTable = true
						end
					else
						useLocalColorTable = true
					end
				else
					if info.globalColorTableFlag then
						globalColorTable = info.globalColorTable
					else
						useLocalColorTable = true
					end
				end
			end # unless
		end # do

		buf = 'GIF89a'
		buf.force_encoding 'ASCII-8BIT'
		buf << [width, height].pack('v2')
		packedFields18 = @gifs[0].packedFields18
		if useLocalColorTable then
			packedFields18 &= ~0x80
		end
		buf << packedFields18
		buf << @gifs[0].backgroundColorIndex
		buf << @gifs[0].pixelAspectRatio
		unless useLocalColorTable then
			buf << globalColorTable
		end

		leftpos = 0
		@gifs.each do|info|
=begin
			# XXX Graphic Control Extension
			buf << [0x21, 0xf9, 0x04].pack('C3')
=end
			# image
			buf << 0x2c
			buf << [leftpos, 0].pack('v2')
			leftpos += info.width
			buf << [info.width, info.height].pack('v2')

			# XXX ignoring original local color table
			packedFields20 = info.packedFields20
			if useLocalColorTable then
				packedFields20 |= 0x80
				packedFields20 &= ~0x07
				packedFields20 |= info.packedFields18 & 0x07
				buf << packedFields20
				buf << info.globalColorTable
			else
				packedFields20 &= ~0x80
				buf << packedFields20
			end
			buf << info.lzwMinimumCodeSize
			buf << info.imageData
		end

		# trailer
		buf << 0x3b

		buf
	end
end

class GifInfo
	# header info
	attr_reader :signature, :version, :screenWidth, :screenHeight
	attr_reader :packedFields18, :globalColorTableFlag, :sizeOfGlobalColorTable, :globalColorTable
	attr_reader :backgroundColorIndex, :pixelAspectRatio

	# image info
	attr_reader :left, :top, :width, :height
	attr_reader :packedFields20, :localColorTableFlag, :sizeOfLocalColorTable, :localColorTable
	attr_reader :lzwMinimumCodeSize, :imageData

	def initialize gifdata
		io = StringIO.new gifdata
		read_header io
		read_body io
	end

	def read_header io
		@signature = io.read 3
		@version = io.read 3
		@screenWidth, @screenHeight = io.read(4).unpack 'v2'
		@packedFields18 = io.readbyte
		@globalColorTableFlag = @packedFields18 & 0x80 != 0
		@backgroundColorIndex = io.readbyte
		@pixelAspectRatio = io.readbyte
		if @globalColorTableFlag then
			@sizeOfGlobalColorTable = 2 ** ((@packedFields18 & 0x07) + 1);
			@globalColorTable = io.read @sizeOfGlobalColorTable * 3
		end

		if DEBUG_DUMP then
			p "Header ========"
			p "signature = #{@signature}"
			p "version = #{@version}"
			p "screenWidth = #{@screenWidth}"
			p "screenHeight = #{@screenHeight}"
			p "packedFields18 = #{@packedFields18}"
			p "globalColorTableFlag = #{@globalColorTableFlag}"
			p "backgroundColorIndex = #{@backgroundColorIndex}"
			p "pixelAspectRatio = #{@pixelAspectRatio}"
			if @globalColorTableFlag then
				p "sizeOfGlobalColorTable = #{@sizeOfGlobalColorTable}"
				p "globalColorTable = #{@globalColorTable}"
			end
			p "========"
		end
	end

	def read_body io
		until io.eof do
			x1 = io.getbyte
			case x1
			when 0x2c then
				read_image io
			when 0x21 then
=begin
				x2 = io.getbyte
				case x2
				when 0xf9 then
				when 0xfe then
				when 0x01 then
				when 0xff then
				else
					raise "unknown x2 byte - 0x%02x" % x2
				end
=end
				raise "Extensions not support"
			when 0x3b then
				trailer
				break
			else
				raise "unknown x1 byte - 0x%02x" % x1
			end
		end
	end

	def read_image io
		@left, @top = io.read(4).unpack 'v2'
		@width, @height = io.read(4).unpack 'v2'
		@packedFields20 = io.readbyte
		@localColorTableFlag = @packedFields20 & 0x80 != 0
		if @localColorTableFlag then
			@sizeOfLocalColorTable = 2 ** ((@packedFields20 & 0x07) + 1)
			@localColorTable = io.read @sizeOfLocalColorTable * 3
		end
		@lzwMinimumCodeSize = io.readbyte

		if DEBUG_DUMP then
			p "Image ========"
			p "left = #{@left}"
			p "top = #{@top}"
			p "width = #{@width}"
			p "height = #{@height}"
			p "packedFields20 = #{@packedFields20}"
			p "localColorTableFlag = #{@localColorTableFlag}"
			if @localColorTableFlag then
				p "sizeOfLocalColorTable = #{@sizeOfLocalColorTable}"
				p "localColorTable = #{@localColorTable}"
			end
			p "lzwMinimumCodeSize = #{@lzwMinimumCodeSize}"
			p "========"
		end

		@imageData = read_subblock io
	end

	def read_subblock io
		buf = ''
		buf.force_encoding 'ASCII-8BIT'
		while (n = io.readbyte) != 0 do
			buf << n
			buf << io.read(n)
		end
		buf << n

		if DEBUG_DUMP then
			p "Sub block ========"
			p "buf = #{buf}"
			p "========"
		end

		buf
	end

	def trailer
		if DEBUG_DUMP then
			p "Trailer ========"
			p "0x3b"
			p "========"
		end
	end
end

=begin
def gifcat filenames
	gifcat = GifCat.new

	filenames.each do|filename|
		gifdata = nil
		open filename, 'rb' do|file|
			file.set_encoding 'ASCII-8BIT'
			gifdata = file.read
		end
		if gifdata and not gifdata.empty? then
			gifcat.concat gifdata
		end
	end

	unless filenames.empty? then
		gifcat.output
	end
end
=end
END_OF_SCRIPT
end

def output_internal_image io
	io.print "Image = [\n"
	first_time = true
	(0 .. 9).each do|i|
		if first_time then
			first_time = false
		else
			io.print ",\n"
		end
		filename = "#{i}.gif"
		open filename do|file|
			file.set_encoding 'ASCII-8BIT'
			until file.eof do
				io.print "\t" + '"'
				(16).times do
					c = file.getbyte
					if c then
						io.print "\\x%02x" % c
					end
				end
				io.print '"'
				unless file.eof then
					io.print " +\n"
				end
			end
		end
	end
	io.print "]\n"
	io.print "Image.each{|s|s.force_encoding 'ASCII-8BIT'}"
end

def output_script_2 io
	io.write <<'END_OF_SCRIPT'
def countup
	n = nil
	open 'counter.txt', 'r+' do|file|
		file.flock File::LOCK_EX
		begin
			n = Integer file.read
			file.pos = 0
			file.write (n + 1).to_s
		ensure
			file.flock File::LOCK_UN
		end
	end
	n
end

def output_gif s
	cat = GifCat.new

	s.each_char do|c|
		n = Integer c
		cat.concat Image[n]
	end

	content = cat.get_content

	print "Content-type: image/gif\r\n"
	print "Content-length: %d\r\n" % content.bytesize
	print "\r\n"
	print content
end

count = countup
count_string = "%06d" % count
output_gif count_string
END_OF_SCRIPT
end

# file check
(0 .. 9).each do|i|
	filename = "#{i}.gif"
	unless FileTest.exist? filename then
		STDERR.print "#{filename} not found\n"
		exit 2
	end
end

open 'counter.cgi', 'w' do|file|
	output_script_1 file
	file.write "\n"
	output_internal_image file
	file.write "\n"
	output_script_2 file
end
